#include "modules/perl/mod_perl.h"

static array_header *Buffer;
static int stderr_log_fd = 0;
static int STDERR_log_fd = 0;

typedef void (*lfunc) (char *);

#define USING_SYSLOG(s) \
   (!stderr_log_fd && strnEQ(s->error_fname, "syslog", 6))

#define unchomp(fd) \
    if(buff[len-1] != '\n') write(fd, "\n", len)

static void lfunc_STDERR(char *buff)
{
    int len = strlen(buff);
    (void)write(STDERR_log_fd, buff, len);
    unchomp(STDERR_log_fd);
}

static void lfunc_stderr(char *buff)
{
    int len = strlen(buff);
    (void)PerlIO_write(PerlIO_stderr(), buff, len);
}

static void lfunc_syslog(char *buff)
{
    syslog(APLOG_ERR, "%s", buff);
}

static void LogSTDERR_flush(void *data)
{
    request_rec *r = (request_rec *)data;
    int i;
    lfunc fptr = USING_SYSLOG(r->server) ?
	lfunc_syslog : lfunc_stderr;
    char **list = (char**)Buffer->elts;

    if(STDERR_log_fd) fptr = lfunc_STDERR;

    for(i = 0; i < Buffer->nelts; i++) {
	(*fptr)(list[i]);
    }
    Buffer->nelts = 0;
}

static int open_logger(server_rec *s, char *logname)
{
    pool *p = perl_get_startup_pool();
    if (*logname == '|') {
        char *pname = server_root_relative(p, (logname+1));
        piped_log *pl = open_piped_log(p, pname);
        if(pl == NULL) croak("can't open piped log `%s'", pname);
        return piped_log_write_fd(pl);
    }
    else {
        char *fname = ap_server_root_relative(p, logname);
        FILE *fp = NULL;
        if (!(fp = ap_pfopen(p, fname, "a"))) {
	    ap_log_error(APLOG_MARK, APLOG_ERR, s, 
		         "httpd: could not open stderr_log file %s.\n", fname);
	    return 0;
	}
	else 
            return fileno(fp);
    }
}

#define dHANDLE(name) GV *handle = gv_fetchpv(name, TRUE, SVt_PVIO)

#define TIEHANDLE(name,obj) \
{ \
      dHANDLE(name); \
      sv_unmagic((SV*)handle, 'q'); \
      sv_magic((SV*)handle, obj, 'q', Nullch, 0); \
}

MODULE = Apache::LogSTDERR    PACKAGE = Apache::LogSTDERR    PREFIX = LogSTDERR_

PROTOTYPES: DISABLE

BOOT:
    if(!Buffer) {
	pool *p = ap_make_sub_pool(NULL);
	Buffer = ap_make_array(p, 100, sizeof(char *));
	av_push(perl_get_av("Apache::LogSTDERR::ISA",1), newSVpv("Apache",6));
    }

int
handler(r)
    Apache r

    PREINIT:
    SV *sv = sv_newmortal();

    CODE:
    sv_setref_pv(sv, "Apache::LogSTDERR", (void*)r);
    TIEHANDLE("STDERR", sv);
    ap_register_cleanup(r->pool, (void*)r, LogSTDERR_flush, ap_null_cleanup);
    RETVAL = OK;

    OUTPUT:
    RETVAL

void
PRINT(r, ...)
    Apache r

    PREINIT:
    int i;

    CODE:
    for(i=1; i<items; i++) {
	char **new;
	STRLEN len;
	char *buf = SvPV(ST(i),len);
	new = (char **)ap_push_array(Buffer);
	*new = ap_pstrndup(r->pool, buf, len);
    }

int
open_stderr_log(s, logname="logs/stderr_log")
    Apache::Server s
    char *logname

    CODE:
    /*if(!USING_SYSLOG(s)) XSRETURN_UNDEF;*/
    RETVAL = stderr_log_fd = open_logger(s, logname);

    OUTPUT:
    RETVAL

int
open_STDERR_log(s, logname="logs/STDERR_log")
    Apache::Server s
    char *logname

    CODE:
    RETVAL = STDERR_log_fd = open_logger(s, logname);

    OUTPUT:
    RETVAL

int
dup_stderr(s)
    Apache::Server s

    CODE:
    if(!stderr_log_fd) XSRETURN_UNDEF;
    fflush(stderr);
    RETVAL = TRUE;
    if (dup2(stderr_log_fd, STDERR_FILENO) == -1) {
	RETVAL = FALSE;
	ap_log_error(APLOG_MARK, APLOG_CRIT, s,
		     "unable to replace stderr with stderr.log");
    }

    OUTPUT:
    RETVAL
